# -*- coding: utf-8 -*-

__version__ = '1.2.1'
__built__ = '1.2.1 dfbfa126d438a7cac30c2a5af8de49629e8a1a94 2021-07-06 09:34 -0600 od'
__author__ = 'Olaf David'
__email__ = 'odavid@colostate.edu'

from .pso import pso
__all__ = ["pso"]

