"""
The :mod:`pyswarms.discrete` module implements various techniques in
discrete optimization. These are techniques that can be applied to a
discrete search-space.
"""

from .binary import BinaryPSO

__all__ = ["BinaryPSO"]
